#ifndef CHESSY_BOARD_H
#define CHESSY_BOARD_H

#include "BitBoard.h"
#include "MoveGeneration.h"

#include <string>

namespace chessy
{
    class Board;
}


class chessy::Board
{
    Bitboard whites[6];
    Bitboard blacks[6];

public:
    Board(void) = default;
    Board(const Board&) = default;
    ~Board(void) = default;
    Board& operator=(const Board&) = default;

    /*!
     * resets the board to the chess starting position.
     */
    void resetBoard(void);

    void setEmpty(void);

private:
    bool tryToMove(Bitboard start, Bitboard end, Bitboard& b);
public:

    inline Bitboard getWhitePawns  (void) const    { return whites[PAWN]; }
    inline Bitboard getWhiteKnights(void) const    { return whites[KNIGHT]; }
    inline Bitboard getWhiteBishops(void) const    { return whites[BISHOP]; }
    inline Bitboard getWhiteRooks  (void) const    { return whites[ROOK]; }
    inline Bitboard getWhiteQueens (void) const    { return whites[QUEEN]; }
    inline Bitboard getWhiteKings  (void) const    { return whites[KING]; }

    inline Bitboard getBlackPawns  (void) const    { return blacks[PAWN]; }
    inline Bitboard getBlackKnights(void) const    { return blacks[KNIGHT]; }
    inline Bitboard getBlackBishops(void) const    { return blacks[BISHOP]; }
    inline Bitboard getBlackRooks  (void) const    { return blacks[ROOK]; }
    inline Bitboard getBlackQueens (void) const    { return blacks[QUEEN]; }
    inline Bitboard getBlackKings  (void) const    { return blacks[KING]; }

    template<Side side>
    inline Bitboard* getBitboards  (void) { return side == WHITE_SIDE ? whites : blacks; }

    inline Bitboard* getBitboards  (Side side) { return side == WHITE_SIDE ? whites : blacks; }

    template<Side side>
    inline const Bitboard* getBitboards(void) const { return side == WHITE_SIDE ? whites : blacks; }

    inline const Bitboard* getBitboards(Side side) const { return side == WHITE_SIDE ? whites : blacks; }


    template<Side side>
    inline Bitboard& getPawns      (void) { return getBitboards<side>()[PAWN]; }
    template<Side side>
    inline Bitboard& getKnights    (void) { return getBitboards<side>()[KNIGHT]; }
    template<Side side>
    inline Bitboard& getBishops    (void) { return getBitboards<side>()[BISHOP]; }
    template<Side side>
    inline Bitboard& getRooks      (void) { return getBitboards<side>()[ROOK]; }
    template<Side side>
    inline Bitboard& getQueens     (void) { return getBitboards<side>()[QUEEN]; }
    template<Side side>
    inline Bitboard& getKing       (void) { return getBitboards<side>()[KING]; }


    template<Side side>
    inline const Bitboard& getPawns(void) const { return getBitboards<side>()[PAWN]; }
    template<Side side>
    inline const Bitboard& getKnights(void) const  { return getBitboards<side>()[KNIGHT]; }
    template<Side side>
    inline const Bitboard& getBishops(void) const  { return getBitboards<side>()[BISHOP]; }
    template<Side side>
    inline const Bitboard& getRooks(void) const  { return getBitboards<side>()[ROOK]; }
    template<Side side>
    inline const Bitboard& getQueens(void) const  { return getBitboards<side>()[QUEEN]; }
    template<Side side>
    inline const Bitboard& getKing (void) const  { return getBitboards<side>()[KING]; }
    template<Side side>
    inline Bitboard get            (void) const
    {
        return  getBitboards<side>()[PAWN] |
                getBitboards<side>()[KNIGHT] |
                getBitboards<side>()[BISHOP] |
                getBitboards<side>()[ROOK] |
                getBitboards<side>()[QUEEN] |
                getBitboards<side>()[KING];
    }

    /*!
     * parses the first part of a FEN string and sets the board
     * to the accourding position
     */
    void setBoard(const std::string& fenPosition);
    std::string getFenBoard(void) const;

    PieceType getWhiteAtPosition(Index i) const;
    PieceType getBlackAtPosition(Index i) const;
    PieceType getAtPosition(Index i) const;

    /*!
     * applies a move to the board
     */
    void move(const Move& move);

    void removeAt(Index index);

    /*!
     * \return
     *      a bitboard where every field occupied by a white piece
     *      is 1, every free field 0
     */
    Bitboard getWhites(void) const;

    /*!
     * \return
     *      a bitboard where every field occupied by a black piece
     *      is 1, every free field 0
     */
    Bitboard getBlacks(void) const;

    /*!
     * \return
     *      a bitboard where every field occupied by any piece
     *      is 1, every free field 0
     */
    Bitboard getOccupied(void) const;

    /*!
     * \return
     *      a bitboard where every field occupied by any piece
     *      is 0, every free field 1
     */
    Bitboard getFree(void) const;

    template<Side side>
    bool isCheck(void) const;
};

#endif // CHESSY_BOARD_H

