#include "MoveGeneration.h"
#include "ChessGame.h"

using namespace chessy;



template<Side side>
typename PawnPushGenerator<side>::MoveIterator PawnPushGenerator<side>::begin(void) const
{
    const Board& board = chessGame.getBoard();
    Bitboard movedPieces = board.getPawns<side>();
    movedPieces.pushOne<side>();
    
    // can't push into occupied squares
    movedPieces &= ~board.getOccupied();
    // use different move generator for promotions
    movedPieces &= Bitboard(~0xFF000000000000FFULL);

    return MoveIterator{ movedPieces };
}


template<Side side>
typename PawnPushGenerator<side>::MoveIterator PawnPushGenerator<side>::end(void) const
{
    return MoveIterator{ 0 };
}


template<Side side>
void PromotionGenerator<side>::MoveIterator::next(void)
{
    if (direction != 1) {
        direction++;
    }
    else if (promotionType != QUEEN) {
        promotionType = static_cast<PieceType>(promotionType + 1);
        direction = -1;
    }
    else {
        promotionType = PieceType::KNIGHT;
        direction = -1;
        ++pawns;
    }
}


template<Side side>
bool PromotionGenerator<side>::MoveIterator::valid(void) const
{
    if (pawns.bitboard == Bitboard{ 0 })
        return true;
    if (direction == 0)
        return !(Bitboard::fromIndex(*pawns +
                (side == WHITE_SIDE ? 8 : -8)) &
                chessGame.getBoard().get<otherSide(side)>());
    else
        return (Bitboard::fromIndex(*pawns +
                (side == WHITE_SIDE ? 8 : -8) + direction) &
                Bitboard(~(Bitboard::aFile | Bitboard::hFile))  &
                chessGame.getBoard().get<otherSide(side)>());
}


template<Side side>
typename PromotionGenerator<side>::MoveIterator PromotionGenerator<side>::begin(void) const
{
    const Board& board = chessGame.getBoard();
    Bitboard pawns = board.getPawns<side>();
    pawns &= Bitboard(0xFF000000000000FFULL);

    return MoveIterator{ chessGame, pawns, PieceType::KNIGHT, -1 };
}


template<Side side>
typename PromotionGenerator<side>::MoveIterator PromotionGenerator<side>::end(void) const
{
    return MoveIterator{ chessGame, 0, PieceType::PAWN, 0 };
}


template<Side side>
typename PawnDoublePushGenerator<side>::MoveIterator
    PawnDoublePushGenerator<side>::begin(void) const
{
    const Board& board = chessGame.getBoard();
    Bitboard movedPieces = side == WHITE_SIDE ?
        board.getWhitePawns() & Bitboard(0xFFULL << 8) :
        board.getBlackPawns() & Bitboard(0xFFULL << 48);
    if (side == WHITE_SIDE)
        movedPieces.moveNorth(2);
    else
        movedPieces.moveSouth(2);

    Bitboard occupied = board.getOccupied();
    occupied |= side == WHITE_SIDE ? occupied.northOne() : occupied.southOne();
    movedPieces &= ~occupied;

    return MoveIterator{ movedPieces };
}


template<Side side>
typename PawnDoublePushGenerator<side>::MoveIterator
    PawnDoublePushGenerator<side>::end(void) const
{
    return MoveIterator{ 0 };
}


template<Side side, Leftright leftright>
typename PawnCaptureGenerator<side, leftright>::MoveIterator
    PawnCaptureGenerator<side, leftright>::begin(void) const
{
    const Board& board = chessGame.getBoard();
    Bitboard movedPieces = side == WHITE_SIDE ?
        board.getWhitePawns() :
        board.getBlackPawns();
    if (side == WHITE_SIDE)
        movedPieces.moveNorthOne();
    else
        movedPieces.moveSouthOne();
    
    if (leftright == LEFT)
        movedPieces.moveEastOne();
    else
        movedPieces.moveWestOne();
    
    movedPieces &= (side == BLACK_SIDE ? board.getWhites() : board.getBlacks());
    movedPieces &= Bitboard(~0xFF000000000000FFULL);

    return MoveIterator{ movedPieces };
}


template<Side side, Leftright leftright>
typename PawnCaptureGenerator<side, leftright>::MoveIterator
    PawnCaptureGenerator<side, leftright>::end(void) const
{
    return MoveIterator{ 0 };
}

const std::array<Bitboard, 64> KnightMoveGenerator::moveSets = generateKnightMoves();

Bitboard KnightMoveGenerator::generateFromIndex(Index i) {
    int row = i.getRow();
    int column = i.getColumn();
    Bitboard result = 0;
    auto makeBitboardIfValid = [](int row, int column) -> Bitboard {
        return row < 8 && column < 8 && row >= 0 && column >= 0 ?
            Bitboard::fromIndex({row, column}) : Bitboard{ 0 };
    };
    result |= makeBitboardIfValid(row - 2, column - 1);
    result |= makeBitboardIfValid(row - 1, column - 2);
    result |= makeBitboardIfValid(row + 2, column - 1);
    result |= makeBitboardIfValid(row + 1, column - 2);
    result |= makeBitboardIfValid(row - 2, column + 1);
    result |= makeBitboardIfValid(row - 1, column + 2);
    result |= makeBitboardIfValid(row + 2, column + 1);
    result |= makeBitboardIfValid(row + 1, column + 2);
    return result;
}


std::array<Bitboard, 64> KnightMoveGenerator::generateKnightMoves(void)
{
    std::array<Bitboard, 64> moves{};
    for (size_t i = 0; i < moves.size(); i++) {
        moves[i] = generateFromIndex(Index{ (int8_t) i });
    }
    return moves;
}


Bitboard PrimitiveQueenMoveGenerator::generateAttackSet(Index i,
    Bitboard enemies, Bitboard friendly, bool straight, bool diagonal)
{
    Bitboard queen = Bitboard::fromIndex(i);
    Bitboard occupied = (enemies | friendly) & ~queen;

    // strange c++ incoming
    auto iterate = [queen, occupied] (void (Bitboard::*moveStep)()) -> Bitboard {
        Bitboard result = 0;
        int a = 0;
        for (Bitboard r = queen; !(r & occupied) && r;) {
            (r.*moveStep)();
            result |= r;
            a++;
        }
        return result;
    };

    Bitboard possibleMoves = 0;
    if (straight)
        possibleMoves |=
            iterate(&Bitboard::moveNorthOne) |
            iterate(&Bitboard::moveSouthOne) |
            iterate(&Bitboard::moveEastOne) |
            iterate(&Bitboard::moveWestOne);
    if (diagonal)
        possibleMoves |=
            iterate(&Bitboard::moveNEOne) |
            iterate(&Bitboard::moveSEOne) |
            iterate(&Bitboard::moveNWOne) |
            iterate(&Bitboard::moveSWOne);   
    return (possibleMoves & ~friendly);
}


Bitboard KingMoveGenerator::generateKingMoves(Bitboard king)
{
    return king.northOne() | king.southOne() | king.eastOne() | king.westOne() |
        king.neOne() | king.nwOne() | king.swOne() | king.seOne();
}


template<Side side>
CastlingGenerator<side>::CastlingGenerator(const ChessGame& game) :
    PositionSet{ 0 }
{
    bool canCastleQueenSide = game.getCanCastleQueenSide(side);
    bool canCastleKingSide = game.getCanCastleKingSide(side);

    if (!canCastleKingSide && !canCastleQueenSide)
        return;

    Bitboard king = game.getBoard().getKing<side>();
    Bitboard targets = 0;
    if (canCastleKingSide) {
        targets |= king.eastOne().eastOne();
    }
    if (canCastleQueenSide) {
        targets |= king.westOne().westOne();
    }
    PositionSet::setBitboard(targets);
}


template<Side side>
void chessy::generatePawnPushes(const ChessGame& cg, std::vector<Move>& moves)
{
    PawnPushGenerator<side> ppg{ cg };
    for (const auto& move : ppg)
        moves.push_back(move);
}


template<Side side>
void chessy::generatePawnDoublePushes(const ChessGame& cg, std::vector<Move>& moves)
{
    PawnDoublePushGenerator<side> pdpg{ cg };
    for (const auto& move : pdpg)
        moves.push_back(move);
}


template<Side side>
void chessy::generatePawnCaptures(const ChessGame& cg, std::vector<Move>& moves)
{
    PawnCaptureGenerator<side, LEFT> pcgl{ cg };
    PawnCaptureGenerator<side, RIGHT> pcgr{ cg };
    for (const auto& move : pcgl)
        moves.push_back(move);
    for (const auto& move : pcgr)
        moves.push_back(move);
}


template<Side side>
void chessy::generatePawnPromotions(const ChessGame& cg, std::vector<Move>& moves)
{
    PromotionGenerator<side> pg{ cg };
    for (const auto& move : pg)
        moves.push_back(move);
}


template<Side side>
void chessy::generateKnightMoves(const ChessGame& cg, std::vector<Move>& moves)
{
    const Board& b = cg.getBoard();
    PositionSet p{ b.getKnights<side>() };
    for (auto pos : p) {
        KnightMoveGenerator g{ pos, b.get<side>() };
        for (auto target : g) {
            moves.emplace_back(pos, target);
        }
    }
}


template<Side side, typename Generator>
void generateMoves(Bitboard position, Bitboard enemies, Bitboard friendly,
    std::vector<Move>& moves)
{
    PositionSet p{ position };
    for (auto pos : p) {
        Generator g{ pos, enemies, friendly };
        for (auto target : g) {
            moves.emplace_back(pos, target);
        }
    }
}


template<Side side>
void chessy::generateQueenMoves(const ChessGame& cg, std::vector<Move>& moves)
{
    const Board& b = cg.getBoard();
    generateMoves<side, PrimitiveQueenMoveGenerator>(
        b.getQueens<side>(), b.get<otherSide(side)>(), b.get<side>(), moves
    );
}


template<Side side>
void chessy::generateRookMoves(const ChessGame& cg, std::vector<Move>& moves)
{
    const Board& b = cg.getBoard();
    generateMoves<side, PrimitiveRookMoveGenerator>(
        b.getRooks<side>(), b.get<otherSide(side)>(), b.get<side>(), moves
    );
}


template<Side side>
void chessy::generateBishopMoves(const ChessGame& cg, std::vector<Move>& moves)
{
    const Board& b = cg.getBoard();
    generateMoves<side,PrimitiveBishopMoveGenerator>(
        b.getBishops<side>(), b.get<otherSide(side)>(), b.get<side>(), moves
    );
}


template<Side side>
void chessy::generateKingMoves(const ChessGame& cg, std::vector<Move>& moves)
{
    const Board& b = cg.getBoard();
    Bitboard king = b.getKing<side>();
    Index kingIndex = king.getLeastSignificantBit();
    for (auto pos : KingMoveGenerator{ king, b.get<side>() })
        moves.emplace_back(kingIndex, pos);
}


template<Side side>
void chessy::generateCastling(const ChessGame& cg, std::vector<Move>& moves)
{
    const Board& b = cg.getBoard();
    Bitboard king = b.getKing<side>();
    Bitboard allOccupied = b.getOccupied();
    if (cg.getCanCastleKingSide(side)) {
        Bitboard target = king.bits >> 2; // move king to the right
        Bitboard rook = king.bits >> 3;
        Bitboard rookTarget = king.bits >> 1;
        if ((rook & b.getRooks<side>()) &&
            !(target & allOccupied) &&
            !(rookTarget & allOccupied)) {
            moves.emplace_back(king.getLeastSignificantBit(),
                target.getLeastSignificantBit(), true);
        }
    }
    if (cg.getCanCastleQueenSide(side)) {
        Bitboard target = king.bits << 2; // move king to the left
        Bitboard rook = king.bits << 4;
        Bitboard rookTarget = king.bits << 1;
        if ((rook & b.getRooks<side>()) &&
            !(target & allOccupied) &&
            !(rookTarget & allOccupied)) {
            moves.emplace_back(king.getLeastSignificantBit(),
                target.getLeastSignificantBit(), true);
        }
    }
}


template<Side side>
void chessy::generateEnPassant(const ChessGame& cg, std::vector<Move>& moves)
{
    if (cg.getEnPassantIndex() != -1) {
        //Bitboard mask = side == WHITE_SIDE ? 0x000000FF00000000;
    }
}

template<Side side>
void chessy::generateAllMoves(const ChessGame& cg, std::vector<Move>& moves)
{
    generatePawnPushes<side>(cg, moves);
    generatePawnDoublePushes<side>(cg, moves);
    generatePawnCaptures<side>(cg, moves);
    generatePawnPromotions<side>(cg, moves);
    generateKnightMoves<side>(cg, moves);
    generateQueenMoves<side>(cg, moves);
    generateBishopMoves<side>(cg, moves);
    generateRookMoves<side>(cg, moves);
    generateKingMoves<side>(cg, moves);
    generateCastling<side>(cg, moves);
    generateEnPassant<side>(cg, moves);
}


template<Side side>
void chessy::orderMoves(const ChessGame& cg, std::vector<Move>& moves)
{
    if (moves.empty())
        return;
    auto a = moves.begin();
    auto b = moves.end() - 1;
    Bitboard enemies = cg.getBoard().get<otherSide(side)>();
    while(a < b) {
        while (a < b && (Bitboard::fromIndex(a->destination) & enemies))
            ++a;
        while (!(Bitboard::fromIndex(b->destination) & enemies) && a < b)
            --b;
        if (a < b)
            std::iter_swap(a, b);
    }
}


// explicit instatiations
namespace chessy
{
    template class PawnPushGenerator<WHITE_SIDE>;
    template class PawnPushGenerator<BLACK_SIDE>;

    template class PromotionGenerator<WHITE_SIDE>;
    template class PromotionGenerator<BLACK_SIDE>;

    template class PawnDoublePushGenerator<WHITE_SIDE>;
    template class PawnDoublePushGenerator<BLACK_SIDE>;

    template class PawnCaptureGenerator<WHITE_SIDE, LEFT>;
    template class PawnCaptureGenerator<WHITE_SIDE, RIGHT>;
    template class PawnCaptureGenerator<BLACK_SIDE, LEFT>;
    template class PawnCaptureGenerator<BLACK_SIDE, RIGHT>;

    template class CastlingGenerator<WHITE_SIDE>;
    template class CastlingGenerator<BLACK_SIDE>;

    template void generateAllMoves<WHITE_SIDE>(const ChessGame&, std::vector<Move>&);
    template void generateAllMoves<BLACK_SIDE>(const ChessGame&, std::vector<Move>&);

    template void orderMoves<WHITE_SIDE>(const ChessGame&, std::vector<Move>&);
    template void orderMoves<BLACK_SIDE>(const ChessGame&, std::vector<Move>&);
}
