#ifndef UCIPARSER_H
#define UCIPARSER_H

#include <iostream>
#include <utility>
#include <vector>
#include <string>
#include <algorithm>
#include <map>
#include <mutex>

#include "ChessGame.h"
#include "TimeManager.h"
#include "Minimax.h"


struct Option
{
    std::string name;
    std::string type;
    std::string defaultValue;
    int min;
    int max;
};


class OptionsManager
{
    std::map<std::string, Option> options = {
        {"", {}}
    };

public:
    OptionsManager(void) = default;

    void setOption(const std::string& name, const Option& option);
};


/*!
 * manages the communication with a UCI connection
 */
class UciParser
{
    using CommandHandler = void(UciParser::*)(const std::vector<std::string>&);
    static const std::map<std::string, CommandHandler> commandHandlers;

    std::istream& in;
    std::ostream& out;
    std::mutex outStreamLock;

    OptionsManager optionsManager;

    bool quitting = false;

    chessy::ChessGame cg;
    //chessy::MiniMax minimax{ cg };

    std::unique_ptr<FixedSearchTimer> fst;
public:
    inline UciParser(std::istream& in = std::cin, std::ostream& out = std::cout) :
        in{ in }, out{ out } {}
    virtual ~UciParser(void) = default;

    virtual int parse(std::istream& in, std::ostream& out);

    virtual int executeLine(const std::string& line);
    virtual int executeCommand(const std::string& command,
                               const std::vector<std::string>& args);

    virtual void sendCommand(const std::string& command,
        const std::vector<std::string>& args = {});

    template<typename T, typename... Args>
    void write(T&& t, Args&&... args)
    {
        std::cout << t << " ";
        write(args...);
    }

    template<typename T>
    inline void write(T&& t)
    {
        std::cout << t << std::endl;
    }

    // command handlers
    virtual void uci            (const std::vector<std::string>& args);
    virtual void debug          (const std::vector<std::string>& args);
    virtual void isready        (const std::vector<std::string>& args);
    virtual void setoption      (const std::vector<std::string>& args);
    virtual void ucinewgame     (const std::vector<std::string>& args);
    virtual void position       (const std::vector<std::string>& args);
    virtual void go             (const std::vector<std::string>& args);
    virtual void stop           (const std::vector<std::string>& args);
    virtual void quit           (const std::vector<std::string>& args);

    // engine-specific extension commands
    virtual void getfen         (const std::vector<std::string>& args);
    virtual void perft          (const std::vector<std::string>& args);

    /*!
     * Is needed for commands that should not trigger any action, but are
     * still valid commands. An example would be the "register" command,
     * since this chess engine does not have such a feature.
     */
    virtual void doNothing      (const std::vector<std::string>& args);
};

#endif // UCIPARSER_H
